import unittest
from datetime import datetime, timedelta, timezone
from itertools import product
from ncplib.helpers import datetime_to_unix_nano, datetime_to_unix, unix_to_datetime, dbm_to_rssi, rssi_to_dbm


TIMESTAMP_VALUES = [
    datetime(1970, 1, 1, 0, 0, tzinfo=timezone.utc),
    datetime.now(tz=timezone.utc),
    datetime(2100, 12, 31, 23, 59, 59, tzinfo=timezone.utc),
]


class DatetimeTestCase(unittest.TestCase):

    def testDatetimeToUnixNano(self):
        for timestamp in TIMESTAMP_VALUES:
            with self.subTest(timestamp=timestamp):
                time, nanotime = datetime_to_unix_nano(timestamp)
                self.assertEqual(unix_to_datetime(time, nanotime, timestamp.tzinfo), timestamp)

    def testDatetimeToUnix(self):
        for timestamp in TIMESTAMP_VALUES:
            timestamp = timestamp - timedelta(microseconds=timestamp.microsecond)  # Round to nearest second.
            with self.subTest(timestamp=timestamp):
                time = datetime_to_unix(timestamp)
                self.assertEqual(unix_to_datetime(time, tz=timestamp.tzinfo), timestamp)


RSSI_VALUES = [0, 10, 2 ** 8 - 1]

REF_LEVEL_VALUES = [-20, 0, 10, 20]


class PowerLevelTestCase(unittest.TestCase):

    def testRssiToDbm(self):
        for rssi, ref_level in product(RSSI_VALUES, REF_LEVEL_VALUES):
            for ref_level in REF_LEVEL_VALUES:
                with self.subTest(rssi=rssi, ref_level=ref_level):
                    self.assertEqual(dbm_to_rssi(rssi_to_dbm(rssi, ref_level), ref_level), rssi)



"""
NCP client
==========

.. currentmodule:: ncplib

:mod:`ncplib` allows you to connect to a :doc:`server` and issue commands.


Overview
--------

Connecting to a NCP server
^^^^^^^^^^^^^^^^^^^^^^^^^^

Connect to a :doc:`server` using :func:`connect`. The returned :class:`Connection` will automatically close when the
connection block exits.

.. code:: python

    import ncplib

    async with await ncplib.connect("127.0.0.1", 9999) as connection:
        pass  # Your client code here.

    # Connection is automatically closed here.


Sending a packet
^^^^^^^^^^^^^^^^

Send a :term:`NCP packet` containing a single :term:`NCP field` using :meth:`Connection.send`:

.. code:: python

    response = connection.send("DSPC", "TIME", SAMP=1024, FCTR=1200)


Receiving replies to a packet
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Connection.send` is a :class:`Response`. Receive a single :term:`NCP field` reply using
:meth:`Response.recv`:

.. code:: python

    field = await response.recv()

Alternatively, use the :class:`Response` as an *async iterator* to loop over multiple :term:`NCP field` replies:

.. code:: python

    async for field in response:
        pass

.. important::
    The *async for loop* will only terminate when the underlying connection closes.


Accessing field data
^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Response.recv` is a :class:`Field`, representing a :term:`NCP field`. Access contained
:term:`NCP parameters <NCP parameter>` using item access:

.. code:: python

    print(field["TSDC"])


Advanced usage
^^^^^^^^^^^^^^

-   :doc:`NCP connection documentation <connection>`.


API reference
-------------

.. autofunction:: connect
"""

import asyncio
import logging
import warnings
from ncplib.connection import Connection
from ncplib.errors import CommandError, CommandWarning


__all__ = (
    "connect",
    "Client",
)


logger = logging.getLogger(__name__)


AUTH_ID = "python3-ncplib"


class Client(Connection):

    def __init__(self, host, port, reader, writer, *, loop, auto_auth, auto_erro, auto_warn, auto_ackn):
        super().__init__(host, port, reader, writer, logger, loop=loop)
        # PacketData handling.
        self._auto_auth = auto_auth
        self._auto_erro = auto_erro
        self._auto_warn = auto_warn
        self._auto_ackn = auto_ackn

    # Connection lifecycle.

    async def _handle_auth(self):
        # Read the initial LINK HELO packet.
        await self.recv_field("LINK", "HELO")
        # Send the connection request.
        self.send("LINK", "CCRE", CIW=AUTH_ID)
        # Read the connection response packet.
        await self.recv_field("LINK", "SCAR")
        # Send the auth request packet.
        self.send("LINK", "CARE", CAR=AUTH_ID)
        # Read the auth response packet.
        await self.recv_field("LINK", "SCON")

    async def _connect(self):
        # Auto-authenticate.
        if self._auto_auth:
            await self._handle_auth()
        # All done.
        self.logger.info("Connected")

    # Receiving fields.

    def _handle_erro(self, field):
        error_detail = field.get("ERRO", None)
        error_code = field.get("ERRC", None)
        if error_detail is not None or error_code is not None:
            raise CommandError(field, error_detail, error_code)
        # Ignore the rest of packet-level errors.
        return field.name != "ERRO"

    def _handle_warn(self, field):
        warning_detail = field.get("WARN", None)
        warning_code = field.get("WARC", None)
        if warning_detail is not None or warning_code is not None:
            warnings.warn(CommandWarning(field, warning_detail, warning_code))
        # Ignore the rest of packet-level warnings.
        return field.name != "WARN"

    def _handle_ackn(self, field):
        return "ACKN" not in field

    def _field_predicate(self, field):
        return (
            # Handle errors.
            (not self._auto_erro or self._handle_erro(field)) and
            # Handle warnings.
            (not self._auto_warn or self._handle_warn(field)) and
            # Handle acks.
            (not self._auto_ackn or self._handle_ackn(field))
        )


async def connect(host, port=9999, *, loop=None, auto_auth=True, auto_erro=True, auto_warn=True, auto_ackn=True):
    """
    Connects to a :doc:`server`.

    This function is a *coroutine*.

    :param str host: The hostname of the :doc:`server`. This can be an IP address or domain name.
    :param int port: The port number of the :doc:`server`.
    :param asyncio.BaseEventLoop loop: The event loop. Defaults to the default asyncio event loop.
    :param bool auto_auth: Automatically perform the :term:`NCP` authentication handshake on connect.
    :param bool auto_erro: Automatically raise a :exc:`CommandError` on receiving an ``ERRO`` :term:`NCP parameter`.
    :param bool auto_warn: Automatically issue a :exc:`CommandWarning` on receiving a ``WARN`` :term:`NCP parameter`.
    :param bool auto_ackn: Automatically ignore :term:`NCP fields <NCP field>` containing an ``ACKN``
        :term:`NCP parameter`.
    :return: The client :class:`Connection`.
    :rtype: Connection
    """
    reader, writer = await asyncio.open_connection(host, port, loop=loop)
    client = Client(
        host,
        port,
        reader,
        writer,
        loop=loop,
        auto_auth=auto_auth,
        auto_erro=auto_erro,
        auto_warn=auto_warn,
        auto_ackn=auto_ackn,
    )
    try:
        await client._connect()
    except:
        client.close()
        await client.wait_closed()
        raise
    return client



"""
NCP server
==========

.. currentmodule:: ncplib

:mod:`ncplib` allows you to create a NCP server and respond to incoming :doc:`client` connections.


Overview
--------

Defining a connection handler
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

A connection handler is a coroutine that starts whenever a new :doc:`client` connects to the server. The provided
:class:`Connection` allows you to receive incoming NCP commands as :class:`Field` instances.

.. code:: python

    async def client_connected(connection):
        pass

When the connection handler exits, the :class:`Connection` will automatically close.


Listening for an incoming packet
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

When writing a :doc:`server`, you most likely want to wait for the connected client to execute a command. Within your
``client_connected`` function, Listen for an incomining :term:`NCP field` using :meth:`Connection.recv`.

.. code:: python

    field = await connection.recv()

Alternatively, use the :class:`Connection` as an *async iterator* to loop over multiple :term:`NCP field` replies:

.. code:: python

    async for field in connection:
        pass

.. important::
    The *async for loop* will only terminate when the underlying connection closes.


Accessing field data
^^^^^^^^^^^^^^^^^^^^

The return value of :meth:`Connection.recv` is a :class:`Field`, representing a :term:`NCP field`.

Access information about the :term:`NCP field` and enclosing :term:`NCP packet`:

.. code:: python

    print(field.packet_type)
    print(field.name)

Access contained :term:`NCP parameters <NCP parameter>` using item access:

.. code:: python

    print(field["FCTR"])


Replying to the incoming field
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Send a reply to an incoming :class:`Field` using :meth:`Field.send`.

.. code:: python

    field.send(ACKN=1)


Putting it all together
^^^^^^^^^^^^^^^^^^^^^^^

A simple ``client_connected`` callback might like this:

.. code:: python

    async def client_connected(connection):
        async for field in connection:
            if field.packet_type == "DSPC" and field.name == "TIME":
                field.send(ACNK=1)
                # Do some more command processing here.
            else:
                field.send(ERRO="Unknown command", ERRC=400)
                break


Start the server
^^^^^^^^^^^^^^^^

Start a new NCP server using :func:`run_app`.

.. code:: python

    from ncplib import run_app

    run_app(client_connected)

:func:`run_app` will block until it receives a ``KeyboardInterrupt``, then shut down the server cleanly and return.


Advanced usage
^^^^^^^^^^^^^^

-   :doc:`NCP connection documentation <connection>`.


API reference
-------------

.. autofunction:: run_app

.. autofunction:: start_server

.. autoclass:: Server
    :members:
"""


import asyncio
import logging
from ncplib.connection import Connection, ClosableContextMixin
from ncplib.errors import DecodeError


__all__ = (
    "run_app",
    "start_server",
    "Server",
)


logger = logging.getLogger(__name__)


class ServerHandler(ClosableContextMixin):

    def __init__(self, client_connected, *, loop, auto_auth):
        self._client_connected = client_connected
        self._loop = loop or asyncio.get_event_loop()
        # Logging.
        self.logger = logger
        # PacketData handling.
        self._auto_auth = auto_auth
        # Active handlers.
        self._handlers = set()

    async def _handle_auth(self, connection):
        connection.send("LINK", "HELO")
        await connection.recv_field("LINK", "CCRE")
        connection.send("LINK", "SCAR")
        await connection.recv_field("LINK", "CARE")
        connection.send("LINK", "SCON")

    async def _handle_client_connected(self, reader, writer):
        remote_host, remote_port = writer.get_extra_info("peername")[:2]
        async with Connection(remote_host, remote_port, reader, writer, self.logger, loop=self._loop) as client:
            try:
                # Handle auth.
                if self._auto_auth:
                    await self._handle_auth(client)
                # Delegate to handler.
                await self._client_connected(client)
            except (asyncio.CancelledError, EOFError, OSError):  # pragma: no cover
                pass
            except DecodeError as ex:
                self.logger.warning("Decode error: {ex}".format(ex=ex))
                client.send("LINK", "ERRO", ERRO="Bad request", ERRC=400)
            except:
                self.logger.exception("Unexpected error")
                client.send("LINK", "ERRO", ERRO="Server error", ERRC=500)

    async def __call__(self, reader, writer):
        handler = self._loop.create_task(self._handle_client_connected(reader, writer))
        self._handlers.add(handler)
        try:
            await handler
        finally:
            self._handlers.remove(handler)

    def close(self):
        for handler in self._handlers:
            handler.cancel()

    async def wait_closed(self):
        if self._handlers:
            await asyncio.gather(*self._handlers, loop=self._loop, return_exceptions=True)


class Server(ClosableContextMixin):

    """
    A :doc:`server`.

    Servers can be used as *async context managers* to automatically shut down the server:

    .. code:: python

        async with server:
            pass

        # Server is automatically shut down.

    .. important::

        Do not instantiate this class directly. Use :func:`start_server` to create a :class:`Server`.
    """

    def __init__(self, handler, server):
        self._handler = handler
        self._server = server

    @property
    def sockets(self):
        return self._server.sockets

    def close(self):
        """
        Shuts down the server.

        After calling this method, use :meth:`wait_closed` to wait for the server to fully shut down.

        .. hint::

            If you use the server as an *async context manager*, there's no need to call :meth:`Server.close`
            manually.
        """
        self._handler.close()
        self._server.close()

    async def wait_closed(self):
        """
        Waits for the server to fully shut down.

        This method is a *coroutine*.

        .. important::

            Only call this method after first calling :meth:`close`.

        .. hint::

            If you use the server as an *async context manager*, there's no need to call
            :meth:`Server.wait_closed` manually.
        """
        await self._handler.wait_closed()
        await self._server.wait_closed()


DEFAULT_HOST = "0.0.0.0"

DEFAULT_PORT = 9999

_start_server_args = """:param callable client_connected: A coroutine function taking a single :class:`Connection`
            argument representing the client connection. When the connection handler exits, the :class:`Connection`
            will automatically close. If the client closes the connection, the connection handler will exit.
    :param str host: The host to bind the server to.
    :param int port: The port to bind the server to.
    :param asyncio.BaseEventLoop loop: The event loop. Defaults to the default asyncio event loop.
    :param bool auto_auth: Automatically perform the :term:`NCP` authentication handshake on client connect.
    """


async def start_server(client_connected, host=DEFAULT_HOST, port=DEFAULT_PORT, *, loop=None, auto_auth=True):
    """
    Creates and returns a new :class:`Server` on the given host and port.

    .. hint::
        Prefer :func:`run_app` unless you need to start multiple servers in parallel.

    """
    loop = loop or asyncio.get_event_loop()
    handler = ServerHandler(client_connected, loop=loop, auto_auth=auto_auth)
    server = await asyncio.start_server(handler, host, port, loop=loop)
    return Server(handler, server)
start_server.__doc__ += _start_server_args + """:return: The created :class:`Server`.
    :rtype: Server
    """


def run_app(client_connected, host=DEFAULT_HOST, port=DEFAULT_PORT, *, loop=None, auto_auth=True):  # pragma: no cover
    """
    Runs a new :doc:`server` on the given host and port.

    This function will block until it receives a ``KeyboardInterrupt``, then shut down the server cleanly and return.

    """
    loop = loop or asyncio.get_event_loop()
    server = loop.run_until_complete(start_server(client_connected, host, port, loop=loop, auto_auth=auto_auth))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        server.close()
        loop.run_until_complete(server.wait_closed())
run_app.__doc__ += _start_server_args



from functools import wraps


def valuedispatch(func):
    registry = {}

    def register(value, func=None):
        def decorator(func):
            registry[value] = func
            return func
        return decorator

    @wraps(func)
    def do_valuedispatch(arg1, *args, **kwargs):
        return registry.get(arg1, func)(arg1, *args, **kwargs)

    do_valuedispatch.registry = registry
    do_valuedispatch.register = register
    return do_valuedispatch



from collections import namedtuple, OrderedDict
from struct import Struct
from ncplib.errors import DecodeError
from ncplib.helpers import unix_to_datetime, datetime_to_unix_nano
from ncplib.values import encode_value, decode_value


# Packet structs.

PACKET_HEADER_STRUCT = Struct("<4s4sIIIII4s")

FIELD_HEADER_STRUCT = Struct("<4s3sBI")

PARAM_HEADER_STRUCT = Struct("<4s3sB")

PACKET_FOOTER_STRUCT = Struct("<I4s")


# Packet constants.

PACKET_HEADER = b"\xdd\xcc\xbb\xaa"

PACKET_FOOTER = b"\xaa\xbb\xcc\xdd"


# Identifier encoding.

def encode_identifier(value):
    try:
        return value.encode(encoding="latin1") + (b"\x00" * (len(value) % 4))
    except UnicodeError:  # pragma: no cover
        raise ValueError("Invalid identifier {}".format(value))


# Identifier decoding.

def decode_identifier(value):
    return value.rstrip(b" \x00").decode(encoding="latin1")


# u24 size encoding.

def encode_u24_size(value):
    return (value // 4).to_bytes(length=4, byteorder="little", signed=False)


# u24 size decoding.

def decode_u24_size(value):
    return int.from_bytes(value, byteorder="little", signed=False) * 4


# Param encoding.

def encode_params(params):
    buf = bytearray()
    for name, value in params.items():
        type_id, encoded_value = encode_value(value)
        size = PARAM_HEADER_STRUCT.size + len(encoded_value)
        padding_size = -size % 4
        buf.extend(PARAM_HEADER_STRUCT.pack(
            encode_identifier(name),
            encode_u24_size(size + padding_size),
            type_id,
        ))
        buf.extend(encoded_value)
        buf.extend(b"\x00" * padding_size)
    return buf


# Param decoding.

def decode_params(buf, offset, limit):
    while offset < limit:
        name, u24_size, type_id = PARAM_HEADER_STRUCT.unpack_from(buf, offset)
        name = decode_identifier(name)
        size = decode_u24_size(u24_size)
        value_encoded = bytes(buf[offset+PARAM_HEADER_STRUCT.size:offset+size])
        value = decode_value(type_id, value_encoded)
        yield name, value
        offset += size
    if offset > limit:  # pragma: no cover
        raise DecodeError("Parameter overflow by {} bytes".format(offset - limit))


# Field encoding.

FieldData = namedtuple("FieldData", ("name", "id", "params",))


def encode_fields(fields):
    buf = bytearray()
    for field in fields:
        encoded_params = encode_params(field.params)
        buf.extend(FIELD_HEADER_STRUCT.pack(
            encode_identifier(field.name),
            encode_u24_size(FIELD_HEADER_STRUCT.size + len(encoded_params)),
            0,  # Field type ID is ignored.
            field.id,
        ))
        buf.extend(encoded_params)
    return buf


# Field decoding.

def decode_fields(buf, offset, limit):
    while offset < limit:
        name, u24_size, type_id, field_id = FIELD_HEADER_STRUCT.unpack_from(buf, offset)
        name = decode_identifier(name)
        size = decode_u24_size(u24_size)
        params = OrderedDict(decode_params(buf, offset+FIELD_HEADER_STRUCT.size, offset+size))
        yield FieldData(
            name=name,
            id=field_id,
            params=params,
        )
        offset += size
    if offset > limit:  # pragma: no cover
        raise DecodeError("Field overflow by {} bytes".format(offset - limit))


# Packet formats.

PACKET_FORMAT_ID = 1


# Packet encoding.

def encode_packet(packet_type, packet_id, timestamp, info, fields):
    encoded_fields = encode_fields(fields)
    # Encode the header.
    buf = bytearray()
    timestamp_unix, timestamp_nano = datetime_to_unix_nano(timestamp)
    buf.extend(PACKET_HEADER_STRUCT.pack(
        PACKET_HEADER,
        encode_identifier(packet_type),
        (PACKET_HEADER_STRUCT.size + len(encoded_fields) + PACKET_FOOTER_STRUCT.size) // 4,
        packet_id,
        PACKET_FORMAT_ID,
        timestamp_unix,
        timestamp_nano,
        info,
    ))
    # Write the packet fields.
    buf.extend(encoded_fields)
    # Encode the packet footer.
    buf.extend(PACKET_FOOTER_STRUCT.pack(
        0,  # No checksum.
        PACKET_FOOTER,
    ))
    # All done!
    return buf


# PacketData decoding.

PacketData = namedtuple("PacketData", ("type", "id", "timestamp", "info", "fields",))


def decode_packet_cps(header_buf):
    (
        header,
        packet_type,
        size_words,
        packet_id,
        format_id,
        time,
        nanotime,
        info,
    ) = PACKET_HEADER_STRUCT.unpack(header_buf)
    packet_type = decode_identifier(packet_type)
    size = size_words * 4
    if header != PACKET_HEADER:  # pragma: no cover
        raise DecodeError("Invalid packet header {}".format(header))
    timestamp = unix_to_datetime(time, nanotime)
    # Decode the rest of the body data.
    size_remaining = size - PACKET_HEADER_STRUCT.size

    def decode_packet_body(body_buf):
        if len(body_buf) > size_remaining:  # pragma: no cover
            raise DecodeError("Packet body overflow by {} bytes".format(len(body_buf) - size_remaining))
        fields = list(decode_fields(body_buf, 0, size_remaining - PACKET_FOOTER_STRUCT.size))
        (
            checksum,
            footer,
        ) = PACKET_FOOTER_STRUCT.unpack_from(body_buf, size_remaining - PACKET_FOOTER_STRUCT.size)
        if footer != PACKET_FOOTER:  # pragma: no cover
            raise DecodeError("Invalid packet footer {}".format(footer))
        # All done!
        return PacketData(
            type=packet_type,
            id=packet_id,
            timestamp=timestamp,
            info=info,
            fields=fields,
        )

    # Return the number of bytes to read, and the function to finish decoding.
    return size_remaining, decode_packet_body


def decode_packet(buf):
    body_size, decode_packet_body = decode_packet_cps(buf[:PACKET_HEADER_STRUCT.size])
    return decode_packet_body(buf[PACKET_HEADER_STRUCT.size:])



from datetime import datetime, timedelta, timezone
from math import floor


__all__ = (
    "datetime_to_unix_nano",
    "datetime_to_unix",
    "unix_to_datetime",
    "dbm_to_rssi",
    "rssi_to_dbm",
)


# Datetimes.

def datetime_to_unix_nano(value_datetime):
    value_datetime = value_datetime.astimezone(timezone.utc)
    return int(floor(value_datetime.timestamp())), int(value_datetime.microsecond * 1000),


def datetime_to_unix(value_datetime):
    return datetime_to_unix_nano(value_datetime)[0]


def unix_to_datetime(value_unix, value_nano=0, tz=timezone.utc):
    return datetime.fromtimestamp(value_unix, tz=tz) + timedelta(microseconds=value_nano // 1000)


# Power levels.

def dbm_to_rssi(value_dbm, ref_level=0.0):
    return int(round((value_dbm - float(ref_level) + 127.5) * 2.0))


def rssi_to_dbm(value_rssi, ref_level=0.0):
    return (float(value_rssi) / 2.0) - 127.5 + float(ref_level)



import unittest
from ncplib import uint


UINT_VALUES = [0, 10, 4294967295]

INVALID_UINT_VALUES = [-1, 4294967296]


class UintTestCase(unittest.TestCase):

    def testValidUint(self):
        for value in UINT_VALUES:
            with self.subTest(value=value):
                self.assertIsInstance(uint(value), uint)

    def testInvalidUint(self):
        for value in INVALID_UINT_VALUES:
            with self.subTest(value=value), self.assertRaises(ValueError):
                uint(value)



import asyncio
import unittest
from functools import wraps


class AsyncTestCase(unittest.TestCase):

    def __init__(self, methodName):
        # Wrap method in coroutine.
        func = getattr(self, methodName)
        if asyncio.iscoroutinefunction(func):
            @wraps(func)
            def do_async_test(*args, **kwargs):
                self.loop.run_until_complete(func(*args, **kwargs))
            setattr(self, methodName, do_async_test)
        # All done!
        super().__init__(methodName)

    # Fixtures.

    def setUp(self):
        super().setUp()
        self.loop = asyncio.new_event_loop()
        self.addCleanup(self.loop.close)



"""
Errors and warnings
===================

.. currentmodule:: ncplib

:term:`NCP` errors and warnings.


API reference
-------------

.. autoexception:: CommandError
    :members:

.. autoexception:: CommandWarning
    :members:

.. autoexception:: DecodeError
    :members:

.. autoexception:: DecodeWarning
    :members:
"""


__all__ = (
    "CommandError",
    "CommandWarning",
    "DecodeError",
    "DecodeWarning",
)


class CommandMixin:

    """
    .. attribute:: field

        The :class:`ncplib.Field` that triggered the error.

    .. attribute:: detail

        The human-readable :class:`str` message from the server.

    .. attribute:: code

        The :class:`int` code from the server,
    """

    def __init__(self, field, detail, code):
        super().__init__("{packet_type} {field_name} '{detail}' (code {code})".format(
            packet_type=field.packet_type,
            field_name=field.name,
            detail=detail,
            code=code,
        ))
        self.field = field
        self.detail = detail
        self.code = code


# Errors.

class CommandError(CommandMixin, Exception):

    """
    Raised by the :doc:`client` when the :doc:`server` sends a :term:`NCP field` containing an ``ERRO`` parameter.

    Can be disabled by setting ``auto_erro`` to :obj:`False` in :func:`ncplib.connect`.
    """
    __doc__ += CommandMixin.__doc__


class DecodeError(Exception):

    """
    Raised when a non-recoverable error was encountered in a :term:`NCP packet`.
    """


# Warnings.

class CommandWarning(CommandMixin, Warning):

    """
    Issued by the :doc:`client` when the :doc:`server` sends a :term:`NCP field` containing a ``WARN`` parameter.

    Can be disabled by setting ``auto_warn`` to :obj:`False` in :func:`ncplib.connect`.
    """
    __doc__ += CommandMixin.__doc__


class DecodeWarning(Warning):

    """
    Issued when a recoverable error was encountered in a :term:`NCP packet`.
    """



"""
Value types
===========

.. currentmodule:: ncplib

Overview
--------

NCP data types are mapped onto python types as follows:

=========== ================================================
NCP type    Python type
=========== ================================================
int32       :class:`int`
uint32      :class:`ncplib.uint`
string      :class:`str`
raw         :class:`bytes`
data int8   :class:`array.array(typecode="b") <array.array>`
data int16  :class:`array.array(typecode="h") <array.array>`
data int32  :class:`array.array(typecode="i") <array.array>`
data uint8  :class:`array.array(typecode="B") <array.array>`
data uint16 :class:`array.array(typecode="H") <array.array>`
data uint32 :class:`array.array(typecode="I") <array.array>`
=========== ================================================


API reference
-------------

.. autoclass:: uint
    :members:
"""


import warnings
from array import array
from functools import singledispatch

from ncplib.errors import DecodeWarning
from ncplib.functional import valuedispatch


__all__ = (
    "uint",
)


# Known type codes.

TYPE_I32 = 0x00

TYPE_U32 = 0x01

TYPE_STRING = 0x02

TYPE_RAW = 0x80

TYPE_ARRAY_U8 = 0x81

TYPE_ARRAY_U16 = 0x82

TYPE_ARRAY_U32 = 0x83

TYPE_ARRAY_I8 = 0x84

TYPE_ARRAY_I16 = 0x85

TYPE_ARRAY_I32 = 0x86


# Type converters.

class uint(int):

    """
    An unsigned integer value.

    Python does not distinguish between signed and unsigned integers, but :term:`NCP` encodes them differently.
    Additionally, the binary encoding restricts the ranges of signed and unsigned integers:

    -   ``-2 ** 31 <= int32 <= 2 ** 31 - 1``
    -   ``0 <= uint32 <= 2 ** 32 - 1``

    To distinguish between ``int32`` and ``uint32`` in :term:`NCP parameter` values, wrap any :class:`int` values to be
    encoded as ``uint32`` in :class:`uint`.
    """

    __slots__ = ()

    def __new__(cls, value):
        if not 0 <= value <= 4294967295:
            raise ValueError("Out of range for unsigned 32 bit integer: {!r}".format(value))
        return super().__new__(cls, value)


# Encoders.

@singledispatch
def encode_value(value):  # pragma: no cover
    raise TypeError("Unsupported value type {}".format(type(value)))


@encode_value.register(int)
def encode_value_int(value):
    return TYPE_I32, value.to_bytes(length=4, byteorder="little", signed=True)


@encode_value.register(uint)
def encode_value_uint(value):
    return TYPE_U32, value.to_bytes(length=4, byteorder="little", signed=False)


@encode_value.register(str)
def encode_value_str(value):
    return TYPE_STRING, value.encode(encoding="utf-8", errors="ignore") + b"\x00"


@encode_value.register(bytes)
@encode_value.register(bytearray)
@encode_value.register(memoryview)
def encode_value_bytes(value):
    return TYPE_RAW, value


ARRAY_TYPE_CODES_TO_TYPE_ID = {
    "B": TYPE_ARRAY_U8,
    "H": TYPE_ARRAY_U16,
    "I": TYPE_ARRAY_U32,
    "b": TYPE_ARRAY_I8,
    "h": TYPE_ARRAY_I16,
    "i": TYPE_ARRAY_I32,
}


@encode_value.register(array)
def encode_value_array(value):
    try:
        return ARRAY_TYPE_CODES_TO_TYPE_ID[value.typecode], value.tobytes()
    except KeyError:  # pragma: no cover
        raise TypeError("Unsupported array type {}".format(value.typecode))


# Decoders.

@valuedispatch
def decode_value(type_id, encoded_value):  # pragma: no cover
    warnings.warn(DecodeWarning("Unsupported type ID", type_id))
    return encoded_value


@decode_value.register(TYPE_I32)
def decode_value_i32(type_id, encoded_value):
    return int.from_bytes(encoded_value, byteorder="little", signed=True)


@decode_value.register(TYPE_U32)
def decode_value_u32(type_id, encoded_value):
    return uint.from_bytes(encoded_value, byteorder="little", signed=False)


@decode_value.register(TYPE_STRING)
def decode_value_string(type_id, encoded_value):
    return encoded_value.split(b"\x00", 1)[0].decode(encoding="utf-8", errors="ignore")


@decode_value.register(TYPE_RAW)
def decode_value_raw(type_id, encoded_value):
    return encoded_value


TYPE_ID_TO_ARRAY_TYPE_CODES = dict(map(reversed, ARRAY_TYPE_CODES_TO_TYPE_ID.items()))


@decode_value.register(TYPE_ARRAY_U8)
@decode_value.register(TYPE_ARRAY_U16)
@decode_value.register(TYPE_ARRAY_U32)
@decode_value.register(TYPE_ARRAY_I8)
@decode_value.register(TYPE_ARRAY_I16)
@decode_value.register(TYPE_ARRAY_I32)
def decode_value_array_u8(type_id, encoded_value):
    return array(TYPE_ID_TO_ARRAY_TYPE_CODES[type_id], encoded_value)



import asyncio
from datetime import datetime
from ncplib import connect, start_server, CommandError, CommandWarning
from tests.base import AsyncTestCase


async def success_server_handler(client):
    async for field in client:
        field.send(ACKN=True)
        field.send(**field)


async def error_server_handler(client):
    raise Exception("BOOM")


class ClientServerTestCase(AsyncTestCase):

    # Helpers.

    async def createServer(self, client_connected=success_server_handler, **kwargs):
        server = await start_server(client_connected, "127.0.0.1", 0, loop=self.loop, **kwargs)
        self.addCleanup(self.loop.run_until_complete, server.wait_closed())
        self.addCleanup(server.close)
        port = server.sockets[0].getsockname()[1]
        client = await connect("127.0.0.1", port, loop=self.loop)
        self.addCleanup(self.loop.run_until_complete, client.wait_closed())
        self.addCleanup(client.close)
        return client

    async def assertMessages(self, response, packet_type, expected_fields):
        fields = {}
        while len(fields) < len(expected_fields):
            field = await response.recv()
            fields[field.name] = field
            # Check the field structure.
            self.assertEqual(field.packet_type, packet_type)
            self.assertIsInstance(field.packet_timestamp, datetime)
            self.assertIn(field.name, expected_fields)
            self.assertIsInstance(field.id, int)
            self.assertEqual(len(field), len(expected_fields[field.name]))
        # Check the field content.
        self.assertEqual(fields, expected_fields)

    # Tests.

    async def testClientTransport(self):
        client = await self.createServer()
        self.assertIsInstance(client.transport, asyncio.WriteTransport)

    async def testSend(self):
        client = await self.createServer()
        response = client.send("PACK", "FIEL", FOO="BAR")
        await self.assertMessages(response, "PACK", {"FIEL": {"FOO": "BAR"}})

    async def testSendFiltersMessages(self):
        client = await self.createServer()
        # Send some junk.
        client.send("PACK", "FIEL", BAZ="QUX")
        # Send a field.
        response = client.send("PACK", "FIEL", FOO="BAR")
        await self.assertMessages(response, "PACK", {"FIEL": {"FOO": "BAR"}})

    async def testSendPacketData(self):
        client = await self.createServer()
        response = client.send_packet("PACK", FIEL={"FOO": "BAR"})
        await self.assertMessages(response, "PACK", {"FIEL": {"FOO": "BAR"}})

    async def testRecvFieldDataConnectionFiltersMessages(self):
        client = await self.createServer()
        client.send("PACK", "JUNK", BAZ="QUX")
        client.send("PACK", "FIEL", FOO="BAR")
        field = await client.recv_field("PACK", "FIEL")
        self.assertEqual(field, {"FOO": "BAR"})

    async def testRecvFieldDataResponseFiltersMessages(self):
        client = await self.createServer()
        client.send("PACK", "FIEL", BAZ="QUX")
        response = client.send("PACK", "FIEL", FOO="BAR")
        field = await response.recv_field("FIEL")
        self.assertEqual(field, {"FOO": "BAR"})

    async def testError(self):
        client = await self.createServer()
        client.send("PACK", "FIEL", ERRO="Boom!", ERRC=10)
        with self.assertRaises(CommandError) as cx:
            await client.recv()
        self.assertEqual(cx.exception.field.packet_type, "PACK")
        self.assertEqual(cx.exception.field.name, "FIEL")
        self.assertEqual(cx.exception.detail, "Boom!")
        self.assertEqual(cx.exception.code, 10)

    async def testWarning(self):
        client = await self.createServer()
        client.send("PACK", "FIEL", WARN="Boom!", WARC=10)
        with self.assertWarns(CommandWarning) as cx:
            await client.recv()
        self.assertEqual(cx.warning.field.packet_type, "PACK")
        self.assertEqual(cx.warning.field.name, "FIEL")
        self.assertEqual(cx.warning.detail, "Boom!")
        self.assertEqual(cx.warning.code, 10)

    async def testEncodeError(self):
        client = await self.createServer()
        client._writer.write(b"Boom!" * 1024)
        client._writer.write_eof()
        with self.assertLogs("ncplib.server", "WARN"):
            with self.assertRaises(CommandError) as cx:
                await client.recv()
        self.assertEqual(cx.exception.field.packet_type, "LINK")
        self.assertEqual(cx.exception.field.name, "ERRO")
        self.assertEqual(cx.exception.detail, "Bad request")
        self.assertEqual(cx.exception.code, 400)

    async def testServerError(self):
        with self.assertLogs("ncplib.server", "ERROR"):
            client = await self.createServer(error_server_handler)
            with self.assertRaises(CommandError) as cx:
                await client.recv()
        self.assertEqual(cx.exception.field.packet_type, "LINK")
        self.assertEqual(cx.exception.field.name, "ERRO")
        self.assertEqual(cx.exception.detail, "Server error")
        self.assertEqual(cx.exception.code, 500)

    async def testServerConnectionError(self):
        with self.assertLogs("ncplib.server", "ERROR"):
            with self.assertRaises(CommandError) as cx:
                await self.createServer(error_server_handler, auto_auth=False)
        self.assertEqual(cx.exception.field.packet_type, "LINK")
        self.assertEqual(cx.exception.field.name, "ERRO")
        self.assertEqual(cx.exception.detail, "Server error")
        self.assertEqual(cx.exception.code, 500)



from setuptools import setup, find_packages

from ncplib import __version__


setup(
    name="ncplib",
    version=".".join(map(str, __version__)),
    license="BSD",
    description="CRFS NCP library for Python 3.",
    author="Dave Hall",
    author_email="dhall@crfs.com",
    url="https://github.com/CRFS/python3-ncplib",
    packages=find_packages(exclude="tests"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.5",
    ],
)



import unittest
from array import array
from datetime import datetime, timezone
from ncplib.packets import decode_packet, encode_packet, PacketData, FieldData
from ncplib import uint


REAL_PACKET = (
    b"\xdd\xcc\xbb\xaaLINK'\x00\x00\x00\x01\x00\x00\x00\x01\x00\x00\x00s\xe9\x8eT(\x05\x1b&4I\xb4\x81HELO\x1d"
    b"\x00\x00\x00\x00\x00\x00\x00NCPV\x0f\x00\x00\x02Beta B01.025:Nov  7 2012, 11:27:52 __TESTING_ONLY__\x00"
    b"SEID\x04\x00\x00\x02monitor\x00MACA\x07\x00\x00\x0200:24:81:b4:49:34\x00\x00\x00\x00\x00\x00\x00\xaa\xbb"
    b"\xcc\xdd"
)


PACKET_VALUES = [
    # Integers.
    (-2 ** 31, -2 ** 31),
    (0, 0),
    (10, 10),
    (2 ** 31 - 1, 2 ** 31 - 1),
    # Unsigned integers.
    (uint(0), uint(0)),
    (uint(10), uint(10)),
    (uint(2 ** 32 - 1), uint(2 ** 32 - 1)),
    # Text.
    ("", ""),
    ("foo", "foo"),
    ("如此这般", "如此这般"),
    # Binary.
    (b"", b""),
    (b"foo", b"foo\x00"),
    # U8 array.
    (array("B", []), array("B", [])),
    (array("B", [0]), array("B", [0, 0, 0, 0])),
    (array("B", [10]), array("B", [10, 0, 0, 0])),
    (array("B", [2 ** 8 - 1]), array("B", [2 ** 8 - 1, 0, 0, 0])),
    # U16 array.
    (array("H", []), array("H", [])),
    (array("H", [0]), array("H", [0, 0])),
    (array("H", [10]), array("H", [10, 0])),
    (array("H", [2 ** 16 - 1]), array("H", [2 ** 16 - 1, 0])),
    # U32 array.
    (array("I", []), array("I", [])),
    (array("I", [0]), array("I", [0])),
    (array("I", [10]), array("I", [10])),
    (array("I", [2 ** 32 - 1]), array("I", [2 ** 32 - 1])),
    # I8 array.
    (array("b", []), array("b", [])),
    (array("b", [-2 ** 7]), array("b", [-2 ** 7, 0, 0, 0])),
    (array("b", [0]), array("b", [0, 0, 0, 0])),
    (array("b", [10]), array("b", [10, 0, 0, 0])),
    (array("b", [2 ** 7 - 1]), array("b", [2 ** 7 - 1, 0, 0, 0])),
    # I16 array.
    (array("h", []), array("h", [])),
    (array("h", [-2 ** 15]), array("h", [-2 ** 15, 0])),
    (array("h", [0]), array("h", [0, 0])),
    (array("h", [10]), array("h", [10, 0])),
    (array("h", [2 ** 15 - 1]), array("h", [2 ** 15 - 1, 0])),
    # I32 array.
    (array("i", []), array("i", [])),
    (array("i", [-2 ** 31]), array("i", [-2 ** 31])),
    (array("i", [0]), array("i", [0])),
    (array("i", [10]), array("i", [10])),
    (array("i", [2 ** 31 - 1]), array("i", [2 ** 31 - 1])),
]


class PacketDatasTestCase(unittest.TestCase):

    def testDecodeRealPacketData(self):
        self.assertEqual(decode_packet(REAL_PACKET).fields, [
            FieldData(name="HELO", id=0, params={
                "NCPV": "Beta B01.025:Nov  7 2012, 11:27:52 __TESTING_ONLY__",
                "SEID": "monitor",
                "MACA": "00:24:81:b4:49:34",
            }),
        ])

    def testEncodeDecodeValue(self):
        packet_timestamp = datetime.now(tz=timezone.utc)
        for value, expected_value in PACKET_VALUES:
            with self.subTest(value=value, expected_value=expected_value):
                expected_packet = PacketData("PACK", 10, packet_timestamp, b"INFO", [
                    FieldData("FIEL", 20, {"PARA": expected_value}),
                ])
                decoded_packet = decode_packet(encode_packet("PACK", 10, packet_timestamp, b"INFO", [
                    FieldData("FIEL", 20, {"PARA": value}),
                ]))
                self.assertEqual(decoded_packet, expected_packet)



"""
NCP connection
==============

.. currentmodule:: ncplib

:term:`NCP` connections are used by the :doc:`client` and :doc:`server` to represent each side of a connection.

Overview
--------

Getting started
^^^^^^^^^^^^^^^

-   :doc:`NCP client documentation <client>`.
-   :doc:`NCP server documentation <client>`.


Multiplexing
^^^^^^^^^^^^

Multiplex multiple parallel commands over a single :class:`Connection` using :mod:`asyncio` task functions:

.. code:: python

    import asyncio

    time_response = connection.send("DSPC", "TIME", SAMP=1024, FCTR=1200)
    conf_response = connection.send("NODE", "CONF", CSTA=1)

    time_response, conf_field = await asyncio.gather(time_response.recv(), conf_response.recv())


Spawning tasks
^^^^^^^^^^^^^^

Spawn a concurrent task to handle long-running commands:

.. code::

    import asyncio

    loop = asyncio.get_event_loop()

    async def handle_dspc_time(field):
        field.send(ACKN=1)
        await asyncio.sleep(10)  # Simulate a blocking task.
        field.send(TSDC=0, TIMM=1)

    for field in connection:
        if field.packet_type == "DSPC" and field.name == "TIME":
            # Spawn a concurrent task to avoid blocking the accept loop.
            loop.create_task(handle_dspc_time(field))
        # Handle other field types here.


API reference
-------------

.. important::

    Do not instantiate these classes directly. Use :func:`connect` to create a :doc:`client` connection.
    Use :func:`start_server` or :func:`run_app` to create a :doc:`server`.

.. autoclass:: Connection
    :members:

.. autoclass:: Response
    :members:

.. autoclass:: Field
    :members:
"""

import asyncio
import logging
from collections.abc import Mapping
from datetime import datetime, timezone
from uuid import getnode as get_mac
from ncplib.errors import DecodeError
from ncplib.packets import FieldData, encode_packet, decode_packet_cps, PACKET_HEADER_STRUCT


__all__ = (
    "Connection",
    "Response",
    "Field",
)


# The last four bytes of the MAC address is used as an ID field.
CLIENT_ID = get_mac().to_bytes(6, "little", signed=False)[-4:]


_send_return_doc = """:return: A :class:`Response` providing access to any :class:`Field` instances received in reply to
            the sent packet.
        :rtype: Response
        :raises ValueError: if any of the packet, field or parameter names were not a valid :term:`identifier`, or any
            of the parameter values were invalid.
        :raises TypeError: if any of the parameter values were not one of the supported
            :doc:`value types <values>`.
        """

_recv_return_doc = """:return: The next :class:`Field` received.
        :rtype: Field
        :raises ncplib.CommandError: if the incoming field contains an ``ERRO`` parameter, and this is a :doc:`client`
            connection with ``auto_erro`` enabled.
        :raises ncplib.DecodeError: if the incoming field was part of an invalid :term:`NCP packet`.
        :raises asyncio.IncompleteReadError: if the connection closed unexpectedly.
        """


class Field(Mapping):

    """
    A :term:`NCP field` received by a :class:`Connection`.

    Access :term:`NCP parameter` values using item access:

    .. code:: python

        print(field["PDAT"])
    """

    def __init__(self, connection, packet, field):
        self._connection = connection
        self._packet = packet
        self._field = field

    @property
    def packet_type(self):
        """
        The type of :term:`NCP packet` that contained this field. This will be a valid :term:`identifier`.
        """
        return self._packet.type

    @property
    def packet_timestamp(self):
        """
        A timezone-aware :class:`datetime.datetime` describing when the containing packet was sent.
        """
        return self._packet.timestamp

    @property
    def name(self):
        """
        The name of the :term:`NCP field`. This will be a valid :term:`identifier`.
        """
        return self._field.name

    @property
    def id(self):
        """
        The unique :class:`int` ID of this field.
        """
        return self._field.id

    def __getitem__(self, name):
        return self._field.params[name]

    def __iter__(self):
        return iter(self._field.params)

    def __len__(self):
        return len(self._field.params)

    def __repr__(self):  # pragma: no cover
        return "<Field {packet_type!r} {field_name!r} {params!r}>".format(
            packet_type=self.packet_type,
            field_name=self.name,
            params=dict(self._field.params.items()),
        )

    def send(self, **params):
        """
        Sends a :term:`NCP packet` containing a single field in reply to this field.

        :param \**params: Keyword arguments, one per :term:`NCP parameter`. Each parameter name should be a valid
            :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        """
        return self._connection._send_packet(self.packet_type, [FieldData(self.name, self.id, params)])
    send.__doc__ += _send_return_doc


class Response:

    """
    A response to a :term:`NCP packet`, returned by :meth:`Connection.send`, :meth:`Connection.send_packet` and
    :meth:`Field.send`.

    Provides access to any :class:`Field` received in reply to the sent packet.

    Responses can be used as *async iterators* to loop over each incoming :class:`Field`:

    .. code:: python

        async for field in response:
            pass

    .. important::
        The *async for loop* will only terminate when the underlying connection closes.
    """

    def __init__(self, connection, predicate, field_list):
        self._connection = connection
        self._predicate = predicate
        self._field_list = field_list

    async def __aiter__(self):  # pragma: no cover
        return self

    async def __anext__(self):
        try:
            return await self.recv()
        except EOFError:
            raise StopAsyncIteration

    async def recv(self):
        """
        Waits for the next :class:`Field` received in reply to the sent :term:`NCP packet`.

        This method is a *coroutine*.

        """
        while True:
            while not self._field_list:
                self._field_list = await self._connection._recv_packet()
            while self._field_list:
                field = self._field_list.pop(0)
                if self._predicate(field):
                    return field
    recv.__doc__ += _recv_return_doc

    async def recv_field(self, field_name):
        """
        Waits for the next matching :class:`Field` received in reply to the sent :term:`NCP packet`.

        This method is a *coroutine*.

        .. hint::

            Prefer :meth:`recv` unless the sent packet contained multiple fields.

        :param str field_name: The field name, must be a valid :term:`identifier`.
        """
        async for field in self:
            if field.name == field_name:
                return field
    recv_field.__doc__ += _recv_return_doc


class ConnectionLoggerAdapter(logging.LoggerAdapter):

    def process(self, msg, kwargs):
        msg, kwargs = super().process(msg, kwargs)
        return "ncp://{host}:{port} - {msg}".format(
            msg=msg,
            **self.extra
        ), kwargs


class ClosableContextMixin:

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        self.close()
        await self.wait_closed()


class Connection(ClosableContextMixin):

    """
    A connection between a :doc:`client` and a :doc:`server`.

    Connections can be used as *async iterators* to loop over each incoming :class:`Field`:

    .. code:: python

        async for field in connection:
            pass

    .. important::
        The *async for loop* will only terminate when the underlying connection closes.

    Connections can also be used as *async context managers* to automatically close the connection:

    .. code:: python

        async with connection:
            pass

        # Connection is automatically closed.

    .. attribute:: logger

        The :class:`logging.Logger` used by this connection. Log messages will be prefixed with the host and port of
        the connection.
    """

    def __init__(self, host, port, reader, writer, logger, *, loop):
        # Logging.
        self._host = host
        self._port = port
        self.logger = ConnectionLoggerAdapter(logger, {
            "host": host,
            "port": port,
        })
        # Packet reading.
        self._reader = reader
        # Packet writing.
        self._writer = writer
        self._id_gen = 0
        # Asyncio.
        self._loop = loop or asyncio.get_event_loop()
        # Multiplexing.
        self._packet_receiver = None

    @property
    def transport(self):
        """
        The :class:`asyncio.WriteTransport` used by this connection.
        """
        return self._writer.transport

    def _gen_id(self):
        self._id_gen += 1
        return self._id_gen

    # Packet reading.

    def _field_predicate(self, field):
        return True

    async def _recv_packet_primary(self):
        try:
            # Decode the packet.
            header_buf = await self._reader.readexactly(PACKET_HEADER_STRUCT.size)
            size_remaining, decode_packet_body = decode_packet_cps(header_buf)
            body_buf = await self._reader.readexactly(size_remaining)
            packet = decode_packet_body(body_buf)
            self.logger.debug("Received packet %s %s", packet.type, packet.fields)
            # All done!
            return list(filter(self._field_predicate, (
                Field(self, packet, field)
                for field
                in packet.fields
            )))
        finally:
            self._packet_receiver = None

    async def _recv_packet(self):
        if self._packet_receiver is None:
            self._packet_receiver = self._loop.create_task(self._recv_packet_primary())
        return (await asyncio.shield(self._packet_receiver, loop=self._loop))

    # Receiving fields.

    async def __aiter__(self):
        return Response(self, lambda field: True, [])

    async def recv(self):
        """
        Waits for the next :class:`Field` received by the connection.

        This method is a *coroutine*.

        """
        return await (await self.__aiter__()).recv()
    recv.__doc__ += _recv_return_doc

    async def recv_field(self, packet_type, field_name):
        """
        Waits for the next matching :class:`Field` received by the connection.

        This method is a *coroutine*.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param str field_name: The field name, must be a valid :term:`identifier`.
        """
        async for field in self:
            if field.packet_type == packet_type and field.name == field_name:
                return field
    recv_field.__doc__ += _recv_return_doc

    # Packet writing.

    def _send_packet(self, packet_type, fields):
        encoded_packet = encode_packet(packet_type, self._gen_id(), datetime.now(tz=timezone.utc), CLIENT_ID, fields)
        self._writer.write(encoded_packet)
        self.logger.debug("Sent packet %s %s", packet_type, fields)
        # Create an iterator of response fields.
        expected_fields = frozenset(
            (field.name, field.id)
            for field
            in fields
        )
        return Response(self, lambda field: (
            field.packet_type == packet_type and
            (field.name, field.id) in expected_fields
        ), [])

    # Sending fields.

    def send_packet(self, packet_type, **fields):
        """
        Sends a :term:`NCP packet` containing multiple :term:`NCP fields <NCP field>`.

        .. hint::

            Prefer :meth:`send` unless you need to send multiple fields in a single packet.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param \**fields: Keyword arguments, one per field. Each field name should be a valid :term:`identifier`, and
            the field value should be a :class:`dict` of parameter names mapped to parameter values. Each parameter name
            should be a valid :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        """
        return self._send_packet(packet_type, [
            FieldData(field_name, self._gen_id(), field_params)
            for field_name, field_params
            in fields.items()
        ])
    send_packet.__doc__ += _send_return_doc

    def send(self, packet_type, field_name, **params):
        """
        Sends a :term:`NCP packet` containing a single :term:`NCP field`.

        :param str packet_type: The packet type, must be a valid :term:`identifier`.
        :param str field_name: The field name, must be a valid :term:`identifier`.
        :param \**params: Keyword arguments, one per :term:`NCP parameter`. Each parameter name should be a valid
            :term:`identifier`, and each parameter value should be one of the supported
            :doc:`value types <values>`.
        """
        return self._send_packet(packet_type, [FieldData(field_name, self._gen_id(), params)])
    send.__doc__ += _send_return_doc

    # Connection lifecycle.

    def close(self):
        """
        Closes the connection.

        After calling this method, use :meth:`wait_closed` to wait for the connection to fully close.

        .. hint::

            If you use the connection as an *async context manager*, there's no need to call :meth:`Connection.close`
            manually.
        """
        # The socket is removed by the transport on connection lost, but there doens't seem to be a documented
        # API for checking this before closing the transport.
        if self.transport._sock is not None:
            try:
                self._writer.write_eof()
                self._writer.close()
            except (EOFError, OSError):  # pragma: no cover
                # If the socket is already closed due to a connection error, we dont' really care.
                pass
        self.logger.info("Closed")

    async def wait_closed(self):
        """
        Waits for the connection to fully close.

        This method is a *coroutine*.

        .. important::

            Only call this method after first calling :meth:`close`.

        .. hint::

            If you use the connection as an *async context manager*, there's no need to call
            :meth:`Connection.wait_closed` manually.
        """
        # Keep reading packets until we get an EOF, meaning that the connection was closed.
        while True:
            try:
                await self._recv_packet()
            except (EOFError, OSError, DecodeError):
                return
